{                                                                           }
{ File:       CLexers.pas                                                   }
{ Function:   Simple C lexer for header conversion                          }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2008 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

unit CLexers;

interface

uses
  Classes;

{ TODO: handle 123i8 and 4567i32 ? }

type
  TTokenType = (ttKeyword, ttIdentifier, ttIntegerLiteral, ttStringLiteral,
    ttOperator, ttPunctuator, ttSingleComment, ttMultiComment, ttEOL, ttEOF);

  // Implement this interface to provide characters for the lexer.
  ICharProvider = interface
  ['{893C6CFA-7E7E-4543-97F5-8015EFA3E6EC}']
    function NextChar: Char;
    function LookAhead: Char;
    function CurrentPosition: Integer;
    function StartPosition: Integer;
    function FullText: string;
  end;

  TCLexer = class(TObject)
  private
  class var
    FKeywords: TStringList;
  var
    FCurrentChar: Char;
    FLastChar: Char;
    FCharProvider: ICharProvider;
    FReplacementBuffer: string;
    FIndex: Integer;
    function GetCurrentPos: Integer;
  protected
    function ReplaceMacro(var Value: string): Boolean;
    function IsKeyword(const S: string): Boolean;
    function DecodeStringLiteral(var Value: string): TTokenType;
    function DecodeComment(var Value: string): TTokenType;
    procedure DecodeHex(var Value: string);
    procedure DecodeOctal(var Value: string);
    function DecodeIdentifier(var Value: string): TTokenType;
    function DecodeIntegerConstant(var Value: string): TTokenType;
    function DecodePunctuator(var Value: string): TTokenType;
    function DecodeOperator(var Value: string): TTokenType;
  public
    function NextChar: Char;
    function LookAhead: Char;
    class function ConvertOperator(const Token: string): string; static;
    class function ConvertCString(S: string): string; static;

    constructor Create(CharProvider: ICharProvider);
    function NextToken(var Token: string): TTokenType;
    class procedure AddKeyword(const S: string); static;
    class procedure AddKeywords(const Keywords: array of string); static;

    property CharProvider: ICharProvider read FCharProvider;
    property CurrentChar: Char read FCurrentChar;
    property LastChar: Char read FLastChar;
    property CurrentPos: Integer read GetCurrentPos;
  end;

const
  LexerEndMarker = Char(-1);

implementation

uses
  SysUtils,
{$IFNDEF UNICODE}
  UnicodeUtils,
{$ENDIF}
  CParserData;

const
  CKeywords: array[1..41] of string = (
    'auto', 'break', 'case', 'char', 'const', 'continue', 'default',
    'do', 'double', 'else', 'enum', 'extern', 'float', 'for', 'goto',
    'if', 'int', 'long', 'register', 'return', 'short', 'signed', 'sizeof',
    'static', 'struct', 'switch', 'typedef', 'union', 'unsigned', 'void',
    'volatile', 'while', '__cdecl', '_cdecl', '__stdcall', '_stdcall',
    '__declspec', '_declspec', '_export', '__export',
    ''
  );

  COperators: array[1..14] of string = (
    '/ div',
    '% mod',
    '& and',
    '| or',
    '^ xor',
    '~ not',
    '&& and',
    '|| or',
    '! not',
    '!= <>',
    '== =',
    '>> shr',
    '<< shl',
    '-> ^.'
  );

{ -- TCLexer -- }

// Perhaps a step by step rewrite and very thorough testing?

// -- Constructor

constructor TCLexer.Create(CharProvider: ICharProvider);
begin
  FCharProvider := CharProvider;
  FCurrentChar := ' ';
  FLastChar := ' ';
  AddKeywords(CKeywords);
end;

// -- Public utility functions

class function TCLexer.ConvertOperator(const Token: string): string;
var
  P: Integer;
  S: string;
begin
  for S in COperators do
  begin
    P := Pos(' ', S);
    if Copy(S, 1, P - 1) = Token then
    begin
      Result := Copy(S, P + 1, MaxInt);
      Exit;
    end;
  end;
  Result := Token;
end;

class function TCLexer.ConvertCString(S: string): string;
var
  First, Last: Integer;
  I: Integer;
  InString: Boolean;

  procedure AddChar(C: Char);
  begin
    if CharInSet(C, [#0..#31, #127..#255]) then
    begin
      if InString then
      begin
        Result := Result + '''';
        InString := False;
      end;
      Result := Result + Format('#%d', [Ord(C)]);
    end
    else
    begin
      if not InString then
      begin
        Result := Result + '''';
        InString := True;
      end;
      Result := Result + C;
    end;
  end;

  function GetOctal(const S: string; I: Integer): Char;
  var
    Value, J: Integer;
  begin
    Value := 0;
    for J := 0 to 2 do
      Value := 8 * Value + Ord(S[I + J]) - Ord('0');
    Result := Char(Value);
  end;

  function GetHex(const S: string; I: Integer): Char;
  var
    Value, J: Integer;
  begin
    Value := 0;
    for J := 1 to 2 do
    begin
      Value := 16 * Value + Ord(S[I + J]);
      case S[I + J] of
        '0'..'9': Dec(Value, Ord('0'));
        'A'..'F': Dec(Value, Ord('A') - 10);
        'a'..'f': Dec(Value, Ord('a') - 10);
      end;
    end;
    Result := Char(Value);
  end;

begin
  Result := '';
  InString := False;
  if Length(S) < 2 then
  begin
    Result := '''' + S + '''';
    Exit;
  end;
  if (S[1] = '"') and (S[Length(S)] = '"') then
  begin
    First := 2;
    Last := Length(S) - 1;
  end
  else
  begin
    First := 1;
    Last := Length(S);
  end;
  I := First;
  repeat
    if S[I] = '\' then
    begin
      Inc(I);
      case S[I] of
        'a': // bell
          AddChar(#7);
        'b': // backspace
          AddChar(#8);
        't': // tab
          AddChar(#9);
        'n': // line feed -- newline
          AddChar(#10);
        'v': // vertical tab
          AddChar(#11);
        'f': // form feed
          AddChar(#12);
        'r': // carriage return -- return
          AddChar(#13);
        '0': // either "\0" or octal e.g. "\077"
          if (I < Length(S)) and CharInSet(S[I + 1], ['0'..'7']) then
          begin
            AddChar(GetOctal(S, I));
            Inc(I, 2);
          end
          else
            AddChar(#0);
        '1'..'3': // octal, e.g. "\377"
          begin
            AddChar(GetOctal(S, I));
            Inc(I, 2);
          end;
        'x': // hex, e.g. "\xFF"
          begin
            AddChar(GetHex(S, I));
            Inc(I, 2);
          end;
        '''': // single quote is doubled
          begin
            AddChar('''');
            AddChar('''');
          end
        else
          AddChar(S[I]);
      end;
    end
    else
      AddChar(S[I]);
    Inc(I);
  until I > Last;
  if Instring then
    Result := Result + '''';
end;

// -- Public functions required to set options

class procedure TCLexer.AddKeyword(const S: string);
begin
  if S <> '' then
    FKeywords.Add(S);
end;

class procedure TCLexer.AddKeywords(const Keywords: array of string);
var
  I: Integer;
begin
  if not Assigned(FKeywords) then
  begin
    FKeywords := TStringList.Create;
    FKeywords.CaseSensitive := True;
    Fkeywords.Duplicates := dupIgnore;
  end;
  for I := Low(Keywords) to High(Keywords) do
    if Keywords[I] <> '' then
      AddKeyword(Keywords[I]);
end;

// On entry: CurrentChar points to first character of constant
// On exit: CurrentChar points to last character of constant
// Value is set to Delphi-formatted string representation of constant
function TCLexer.DecodeIntegerConstant(var Value: string): TTokenType;
begin
  Result := ttIntegerLiteral;
  Value := '';
  if CurrentChar = '0' then
  begin
    if CharInSet(LookAhead, ['X', 'x']) then
      DecodeHex(Value)
    else
      DecodeOctal(Value);
  end
  else
  begin
    Value := Value + CurrentChar;
    while CharInSet(LookAhead, ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']) do
    begin
      NextChar;
      Value := Value + CurrentChar;
    end;
  end;
  if CharInSet(LookAhead, ['U', 'u', 'L', 'l']) then
    NextChar;
end;

// On entry: CurrentChar points to first character (0) of constant
// On exit: CurrentChar points to last character of constant
// Value is set to hex Delphi-formatted string representation of constant
procedure TCLexer.DecodeHex(var Value: string);
begin
  NextChar;
  Value := '$';
  while CharInSet(LookAhead, ['0'..'9', 'A'..'F', 'a'..'f'])  do
  begin
    NextChar;
    Value := Value + UpCase(CurrentChar);
  end;
end;

// On entry: CurrentChar points to first character (0) of constant
// On exit: CurrentChar points to last character of constant
// Value is set to decimal Delphi-formatted string representation of constant
procedure TCLexer.DecodeOctal(var Value: string);
var
  IntValue: Integer;
begin
  IntValue := 0;
  while CharInSet(LookAhead, ['0'..'7']) do
  begin
    IntValue := 8 * IntValue + Ord(LookAhead) - Ord('0');
    NextChar;
  end;
  Value := IntToStr(IntValue);
end;

// On entry: CurrentChar points to second character of comment (/ or *)
// On exit: CurrentChar points to last character of comment
// Value is set to Delphi-formatted version of comment
function TCLexer.DecodeComment(var Value: string): TTokenType;
begin
  Value := '';
  if CurrentChar = '/' then
  begin
    Result := ttSingleComment;
    while (LookAhead <> #13) and (LookAhead <> LexerEndMarker) do
    begin
      NextChar;
      Value := Value + CurrentChar;
    end;
    Value := '//' + Value;
  end
  else
  begin
    Result := ttMultiComment;
    NextChar;
    NextChar;
    while not ((LastChar = '*') and (CurrentChar = '/')) do
    begin
      Value := Value + LastChar;
      NextChar;
    end;
    Value := '{' + Value + '}'
  end;
end;

{ TODO: I'd like to expand or ignore macros with parameters, e.g.

          __in_bcount(dwSigLen)   CONST BYTE  *pbSignature,

        The lexer should be able to parse known names and see if it
        has any params, notified in the list like:

          __in_bcount(

        The lexer would only have to read over the expression
        in parentheses.

        Note: One could perhaps even ignore __declspec like that, and remove
        the hardwired code in the parser.
}

// Tricky -- if a replacement is done, the replacement is
//           placed in an internal buffer and NextChar
//           first reads from this buffer (until it reaches the end)
//           before continuing to use CharProvider.NextChar.

function TCLexer.ReplaceMacro(var Value: string): Boolean;
var
  Replacement: string;

  // Recursively replaces macros with their definitions
  // IOW:
  //   WINOLEAPI -> STDAPI -> HRESULT __stdcall -> HResult __stdcall
  function Split(const S: string): string;
  var
    List: TStringList;
    I: Integer;
  begin
    List := TStringList.Create;
    List.CaseSensitive := True;
    List.QuoteChar := '@';
    List.StrictDelimiter := True;
    List.Delimiter := ' ';
    try
      List.DelimitedText := S;
      for I := 0 to List.Count - 1 do
        if TCParserData.Macros.Values[List[I]] <> '' then
          List[I] := Split(TCParserData.Macros.Values[List[I]]);
      Result := '';
      for I := 0 to List.Count - 1 do
        if List[I] <> '#' then
          Result := Result + ' ' + List[I];
    finally
      List.Free;
    end;
  end;

begin
  Result := False;
  Replacement := TCParserData.Macros.Values[Value];
  if Replacement <> '' then
  begin
    Result := True;
    Replacement := Split(Replacement);
    if FIndex = 0 then
      Inc(FIndex);
    if Replacement <> '#' then
    begin
      if FReplacementBuffer <> '' then
        FReplacementBuffer := FReplacementBuffer + ' ' + Replacement + ' '
      else
        FReplacementBuffer := Replacement + ' ';
    end;
    FLastChar := ' ';
    FCurrentChar := ' ';
    Value := '';
  end;
end;

// On entry: CurrentChar points to first character of identifier
// On exit: CurrentChar points to last character of identifier
// Value is set to identifier
function TCLexer.DecodeIdentifier(var Value: string): TTokenType;
begin
  Value := CurrentChar;
  while CharInSet(LookAhead, ['A'..'Z', 'a'..'z', '_', '0'..'9']) do
  begin
     NextChar;
     Value := Value + CurrentChar;
  end;
  if ReplaceMacro(Value) then
  begin
    Result := NextToken(Value);
  end
  else if IsKeyword(Value) then
    Result := ttKeyword
  else
    Result := ttIdentifier;
end;

{
  C Operators:
    + - / % * & | ^ ~ ! > <
    && || += -= /= *= %= &= |= ^= != == >= <= ++ -- << >> ->
    >>= <<=
}

const // Do not localize!
  CDoubleOps = '&& || += -= /= *= %= &= |= ^= != == >= <= >> << ->';

// On entry: CurrentChar points to first character of operator
// On exit: CurrentChar points to last charcater of operator
// Value is set to Delphi equivalent of operator (if any)
function TCLexer.DecodeOperator(var Value: string): TTokenType;
begin
  Result := ttOperator;
  Value := CurrentChar;

  if (LookAhead <> ' ') and
     (Pos(CurrentChar + LookAhead, CDoubleOps) > 0) then
  begin
    NextChar;
    Value := Value + CurrentChar;
  end;

  if ((Value = '<<') or (Value = '>>')) and (LookAhead = '=') then
  begin
    NextChar;
    Value := Value + CurrentChar;
  end;
end;

// On entry: CurrentChar points to punctuator
// On exit: CurrentChar points to punctuator
// Value is set to punctuator
function TCLexer.DecodePunctuator(var Value: string): TTokenType;
begin
  Result := ttPunctuator;
  Value := CurrentChar;
end;

// On entry: CurrentChar points past start quote of string
// On exit: CurrentChar points to closing quote of string
// Value is set to Delphi-formatted string
function TCLexer.DecodeStringLiteral(var Value: string): TTokenType;
begin
  Result := ttStringLiteral;
  Value := '';
  NextChar;
  while CurrentChar <> '"' do
  begin
    Value := Value + CurrentChar;
    NextChar;
  end;
  if CurrentChar <> '"' then
    raise Exception.Create('" expected');
  Value := ConvertCString('"' + Value + '"');
end;

function TCLexer.GetCurrentPos: Integer;
begin
  Result := FCharProvider.CurrentPosition;
end;

function TCLexer.NextToken(var Token: string): TTokenType;
begin
  NextChar;
  while CharInSet(CurrentChar, [#0..#12, #14..' ']) do
  begin
    NextChar;
    FLastChar := ' ';
  end;
  case CurrentChar of
    #13:
      begin
        Result := ttEOL;
        Token := '~EOL';
      end;
    LexerEndMarker:
      begin
        Result := ttEOF;
        Token := '~EOF';
      end;
    '0'..'9':
      Result := DecodeIntegerConstant(Token);
    '_', 'a'..'z', 'A'..'Z':
      begin
        if (CurrentChar = 'L') and (LookAhead = '"') then
        begin
          NextChar; // Read past 'L'
          Result := DecodeStringLiteral(Token);
        end
        else
          Result := DecodeIdentifier(Token);
      end;
    '>', '<', '=', '+', '-', '^', '&', '|', '~', '?', ':', '.', '*':
      Result := DecodeOperator(Token);
    '"':
      Result := DecodeStringLiteral(Token);
    '/':
      begin
        if CharInSet(LookAhead, ['/', '*']) then

        // TODO: why NextChar here?
        begin
          NextChar;
          Result := DecodeComment(Token);
        end
        else
          Result := DecodeOperator(Token);
      end;
    else
      Result := DecodePunctuator(Token);
  end;
end;

function TCLexer.IsKeyword(const S: string): Boolean;
begin
  Result := FKeywords.IndexOf(S) >= 0;
end;

// Private functions required for main loop.

function TCLexer.NextChar: Char;
begin
  FLastChar := FCurrentChar;
  // If we have text in the (macro) replacement buffer, read from that...
  if (FReplacementBuffer <> '') and (FIndex <= Length(FReplacementBuffer)) then
  begin
    Inc(FIndex);
    if FIndex <= Length(FReplacementBuffer) then
      Result := FReplacementBuffer[FIndex]
    else
    begin
      FReplacementBuffer := '';
      FIndex := 0;
      Result := FCharProvider.NextChar;
    end;
  end
  else
  // ... otherwise read from the source.
    Result := FCharProvider.NextChar;
  FCurrentChar := Result;
end;

function TCLexer.LookAhead: Char;
begin
  if (FReplacementBuffer <> '') and (FIndex < Length(FReplacementBuffer)) then
    Result := FReplacementBuffer[FIndex + 1]
  else
    Result := CharProvider.LookAhead;
end;

end.

