{                                                                           }
{ File:       ConvertExpertImpl.pas                                         }
{ Function:   Implementation of actions for API Conversion Package          }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2005 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

unit ConvertExpertImpl;

interface

uses ToolsAPI, Classes, CLexers, CParsers, CParserData;

type
  CToDelphiConverter = class
  public

    // Add $EXTERNALSYM for current symbol, on the line above the symbol.
    // e.g.
    //
    //   {$EXTERNALSYM AddExternalSym}
    //   procedure AddExternalSym;
    //
    class procedure InsertExternalSym; static;

    // Add "{$HPPEMIT '" and "'}" around the current source line.
    // e.g.
    //
    //  {$HPPEMIT '#include "windows.h"'}
    //
    class procedure ConvertToHppEmit; static;

    // Add the header as proposed by Borland, but slightly
    // altered (cleaned up) by me.
    //
    class procedure AddJediHeader; static;

    // Swap C style to Pascal style declaration of fields or parameters.
    // e.g.
    //
    //   "LPCTSTR Alpha;"  --> "Alpha: LPCTSTR;"
    //
    // Commas, like in C style parameter lists, are converted to
    // semicolons:
    //
    // "LPCTSTR Alpha, LPCTSTR Beta," --> "Alpha: LPCTSTR; Beta: LPCTSTR;"
    //
    class procedure SwapDeclarations; static;

    // Convert an entire typedef to Delphi.
    //
    // This conversion converts certain typedefs to Delphi. It currently
    // converts simple typedefs, typedefs of structs and unions, and
    // typedefs of enums. It will generate the entire translation, taking
    // tags, array declarations, pointer declarations, etc. into account.
    //
    // This is the first converter which uses my new parser. This parser
    // is found in CParsers and uses the lexer in CLexers to tokenize.
    //
    class procedure ConvertTypedef; static;

    // Convert a #define macro to Delphi.
    //
    // This will convert simple #defines that define constant values as
    // well as macro functions. The latter only convert operators to
    // their Delphi equivalent and generates some kind of function body.
    //
    // Example:
    //
    //   #define GET_ALG_CLASS(x)    (x & (7 << 13))
    //
    // ConvertDefine generates:
    //
    //   {$EXTERNALSYM GET_ALG_CLASS}
    //   function GET_ALG_CLASS(x: DWORD): DWORD;
    //   (*
    //   begin
    //     Result := (x and (7 shl 13));
    //   end;
    //   *)
    //
    class procedure ConvertDefine; static;

    // Convert an entire function declaration to Delphi.
    //
    // This knows about calling conventions and MS's __declspec keyword.
    // It will ignore any leading extern "C" or extern "C++" specifiers.
    //
    // Uses the same parser as ConvertTypedef and ConvertDefine.
    //
    class procedure ConvertFunction; static;
  private
    class procedure WriteSimpleText(Text: string;
      Index, StartPos: Integer); static;
    class procedure WriteParsedOutput(Provider: ICharProvider;
      const Text: string); static;
    class procedure CommentOut(var S: string); static;
    class procedure UpdateCursorPos(Editor: IOTASourceEditor);
  end;

implementation

uses
  {$IFDEF UNICODE}StrHlpr,{$ENDIF}
  {$IFNDEF UNICODE}UnicodeUtils,{$ENDIF}
  Windows, SysUtils, Menus, ConvertExpertModule, TextMan,
  EditorTools, Dialogs, StrUtils, ConvertConsts, ConvertUtils, ConvertSettings,
  IniFiles, ConvertHeaderDlg, Controls, ValEdit;

type
  TypeModifier = (tmNone, tmUnsigned, tmSigned, tmLong, tmShort, tmOther);
  TypeModifiers = set of TypeModifier;
  TypeKind = (ctUnknown, ctInt, ctChar, ctFloat, ctDouble, ctVoid);

const
  CRLF = #13#10;
  SJediHeader =
    '{*****************************************************************}' + CRLF +
    '{                                                                 }' + CRLF +
    '{            CodeGear Delphi Runtime Library                      }' + CRLF +
    '{            %-52s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{            Coypright (c) %-38s }' + CRLF +
    '{            Converted %-42s }' + CRLF +
    '{            Last modified %-38s }' + CRLF +
    '{            Version %-44s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{*****************************************************************}' + CRLF +
    CRLF +

    // Here comes the legalese.
    '{*****************************************************************}' + CRLF +
    '{                                                                 }' + CRLF +
    '{ The contents of this file are subject to the Mozilla Public     }' + CRLF +
    '{ License Version 1.1 (the "License"). you may not use this file  }' + CRLF +
    '{ except in compliance with the License. You may obtain a copy of }' + CRLF +
    '{ the License at http://www.mozilla.org/MPL/MPL-1.1.html          }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Software distributed under the License is distributed on an     }' + CRLF +
    '{ "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or  }' + CRLF +
    '{ implied. See the License for the specific language governing    }' + CRLF +
    '{ rights and limitations under the License.                       }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ The original code is: %-41s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ The initial developer of the original translation is            }' + CRLF +
    '{ %-63s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Portions created by %-43s }' + CRLF +
    '{ Copyright (C) %-49s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Portions created by %-43s }' + CRLF +
    '{ Copyright (C) %-49s }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ All Rights Reserved.                                            }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Contributor(s):                                                 }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Notes:                                                          }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Modification history:                                           }' + CRLF +
    '{                                                                 }' + CRLF +
    '{ Known Issues:                                                   }' + CRLF +
    '{                                                                 }' + CRLF +
    '{*****************************************************************}' + CRLF;

  CMonths: array[1..12] of string = (
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
  );

type
  TCharProvider = class(TInterfacedObject, ICharProvider)
  private
    FStringPos: Integer;
    FStartPos: Integer;
    FReadPos: Integer;
    FFullText: string;
    function NextBuffer: Boolean;
  public
    procedure Back;
    function CurrentPosition: Integer;
    function LookAhead: Char;
    function NextChar: Char;
    function StartPosition: Integer;
    constructor Create(AtCurrentPos: Boolean = False);
    function FullText: string;
  end;

class procedure CToDelphiConverter.AddJediHeader;
var
  SourceEditor: IOTASourceEditor;
  Writer: IOTAEditWriter;
  HeaderDlg: THeaderDlg;
  HeaderText: string;
  Year, Month, Date: Word;
  L: TStringList;
begin
  L := TStringList.Create;
  try
    HeaderDlg := THeaderDlg.Create(nil);
    try
      if IsPositiveResult(HeaderDlg.ShowModal) then
        L.Assign(HeaderDlg.HeaderValueList.Strings);
    finally
      HeaderDlg.Free;
    end;

    if L.Count = 8 then
    begin
      DecodeDate(Now, Year, Month, Date);
      HeaderText := Format(SJediHeader,
        [
          L.Values['Unit name'] + ' interface unit',
          L.Values['Copyright year'] + ' ' + L.Values['Copyright holder'],
          Format('%d %s %d %s', [Date, CMonths[Month], Year, L.Values['Copyright holder']]),
          Format('%d %s %d %s', [Date, CMonths[Month], Year, L.Values['Copyright holder']]),
          L.Values['Version number'],
          L.Values['Unit name'] + '.h, released ' + L.Values['Original copyright year'] + '.',
          L.Values['Developer'] + ' (' + L.Values['Developer email'] + ').',
          L.Values['Copyright holder'] + ' are',
          L.Values['Copyright year'] + ' ' + L.Values['Copyright holder'] + '.',
          L.Values['Original copyright holder'] + ' are',
          L.Values['Original copyright year'] + ' ' +  L.Values['Original copyright holder'] + '.'
        ]);
      SourceEditor := GetEditor;
      Writer := SourceEditor.CreateUndoableWriter;
      Writer.Insert(PAnsiChar(UTF8String(HeaderText)));
    end;
  finally
    L.Free;
  end;
end;

// TODO: use TCParser and TCLexer

class procedure CToDelphiConverter.InsertExternalSym;
var
  Editor: ToolsAPI.IOTASourceEditor;
  BufferStart, Index: LongInt;
  CurLine: string;
  Writer: IOTAEditWriter;
  Indent: string;
  Symbol: string;
begin
  // Get current source editor
  Editor := GetEditor;

  // Get current line and character index in source editor.
  CurLine := GetCurrentLine(Editor, BufferStart, Index);

  // Get leading white space (indentation)
  Indent := LeadingWhiteSpace(CurLine);

  // If Indent length >= line length, then line is empty.
  if Length(Indent) < Length(CurLine) then
  begin
    Symbol := TextMan.IdentifierAtIndex(CurLine, Index);
    if Symbol <> '' then
    begin
      while CharInSet(Symbol[Length(Symbol)], [':', '=', '(', ',']) do
        Delete(Symbol, Length(Symbol), 1);

      // Insert string to editor at position buffer start.
      Writer := Editor.CreateUndoableWriter;
      Writer.CopyTo(BufferStart);
      Writer.Insert(PAnsiChar(UTF8String(Format('%s{$EXTERNALSYM %s}' + CRLF, // Do not translate!
        [Indent, Symbol]))));
      Writer := nil;
      UpdateCursorPos(Editor);
    end;
  end;
end;

class procedure CToDelphiConverter.ConvertToHppEmit;
var
  Editor: IOTASourceEditor;
  CurLine: string;
  BufferStart, Index: LongInt;
  Writer: IOTAEditWriter;
  CharPos: TOTACharPos;
begin
  Editor := GetEditor;

  CurLine := GetCurrentLine(Editor, BufferStart, Index);
  CurLine := TrimRight(CurLine);

  Writer := Editor.CreateUndoableWriter;
  Writer.CopyTo(BufferStart);
  Writer.DeleteTo(BufferStart + Length(CurLine));
  Writer.Insert(PAnsiChar(UTF8String('{$HPPEMIT ''' + CurLine + '''}'))); // Do not translate!
  Writer := nil;

  CharPos := Editor.EditViews[0].PosToCharPos(BufferStart);
  EditorTools.NextLine(Editor, CharPos);
end;

{ Routines using TCParser }

class procedure CToDelphiConverter.CommentOut(var S: string);
var
  L: TStringList;
  I: Integer;
begin
  L := TStringList.Create;
  try
    L.Text := S;
    for I := 0 to L.Count - 1 do
      L[I] := '////' + L[I];
    S := L.Text;
  finally
    L.Free;
  end;
end;

class procedure CToDelphiConverter.UpdateCursorPos(Editor: IOTASourceEditor);
var
  ViewSize: TSize;
  CharPos: TOTACharPos;
  CursorPos: TOTAEditPos;
  EditView: IOTAEditView;
begin
  EditView := Editor.EditViews[0];
  ViewSize := EditView.ViewSize;
  // Get the cursor pos of that view
  CursorPos := EditView.CursorPos;
  Inc(CursorPos.Line);
  EditView.CursorPos := CursorPos;
  //Convert the new pos to a CharPos.
  EditView.ConvertPos(True, Cursorpos, CharPos);
  // Only scroll if cursor would go off Screen.
  if CursorPos.Line >= EditView.TopPos.Line + ViewSize.cy - 4 then
  begin
    // Scroll up.
    CursorPos := EditView.TopPos;
    Inc(CursorPos.Line, 2);
    // don't know why 2.
    EditView.TopPos := CursorPos;
  end;
  // Update cursor on screen.
  Editor.BlockStart := CharPos;
  Editor.BlockAfter := CharPos;
end;

class procedure CToDelphiConverter.WriteParsedOutput(
  Provider: ICharProvider; const Text: string);
var
  Editor: IOTASourceEditor;
  Writer: IOTAEditWriter;
begin
  Editor := GetEditor;
  Writer := Editor.CreateUndoableWriter;
  Writer.CopyTo(Provider.StartPosition);
  Writer.DeleteTo(Provider.CurrentPosition);
  Writer.Insert(PAnsiChar(UTF8String(Text)));
end;

class procedure CToDelphiConverter.ConvertTypedef;
var
  ParserOutput: string;
  ParsedText: string;
  Parser: TCParser;
  Provider: ICharProvider;
begin
  Provider := TCharProvider.Create;
  Parser := TCParser.Create(Provider);
  try
    try
      Parser.ParseTypedef;
      ParserOutput := Parser.Output;
    except
      on E: Exception do
      begin
        ShowMessage(E.Message);
        Exit;
      end;
    end;
  finally
    Parser.Free;
  end;

  if poKeep in TCParserData.Options then
  begin
    ParsedText := Provider.FullText;
    CommentOut(ParsedText);
    ParsedText := ParsedText + CRLF + ParserOutput;
  end
  else
    ParsedText := ParserOutput;
  
  WriteParsedOutput(Provider, ParsedText);
end;

class procedure CToDelphiConverter.ConvertDefine;
var
  Provider: ICharProvider;
  Parser: TCParser;
  ParserOutput: string;
  Editor: IOTASourceEditor;
  Pos: TOTACharPos;
begin
  Provider := TCharProvider.Create;
  Parser := TCParser.Create(Provider);
  try
    try
      Parser.ParseDefine;
      ParserOutput := Parser.Output;
    except
      on E: Exception do
      begin
        ShowMessage(E.Message);
        Exit;
      end;
    end;
  finally
    Parser.Free;
  end;

  WriteParsedOutput(Provider, ParserOutput);

  Editor := GetEditor;
  Pos := Editor.EditViews[0].PosToCharPos(Provider.StartPosition);
  EditorTools.NextLine(Editor, Pos);
  UpdateCursorPos(Editor);
end;

class procedure CToDelphiConverter.ConvertFunction;
var
  Provider: ICharProvider;
  Parser: TCParser;
  ParserOutput: string;
  ParsedText: string;
begin
  Provider := TCharProvider.Create;
  Parser := TCParser.Create(Provider);
  try
    try
      Parser.ParseFunctionDeclaration;
      ParserOutput := Parser.Output;
    except
      on E: Exception do
      begin
        ShowMessage(E.Message);
        Exit;
      end;
    end;
  finally
    Parser.Free;
  end;

  if poKeep in TCParserData.Options then
  begin
    ParsedText := Provider.FullText;
    CommentOut(ParsedText);
    WriteParsedOutput(Provider, ParsedText + CRLF + ParserOutput);
  end
  else
    WriteParsedOutput(Provider, ParserOutput);
end;

class procedure CToDelphiConverter.WriteSimpleText(Text: string; Index: Integer; StartPos: Integer);
var
  Writer: IOTAEditWriter;
begin
  Writer := GetEditor.CreateUndoableWriter;
  Writer.CopyTo(StartPos);
  Writer.DeleteTo(StartPos + Index);
  Writer.Insert(PAnsiChar(UTF8String(Text)));
end;

class procedure CToDelphiConverter.SwapDeclarations;
var
  Provider: ICharProvider;
  Parser: TCParser;
  ParserOutput: string;
  StartPos: Integer;
  Text: string;
  Index: Integer;
begin
  Provider := TCharProvider.Create(True);
  Parser := TCParser.Create(Provider);
  try
    try
      Parser.ParseSimpleDeclaration;
      ParserOutput := Parser.Output;
    except
      on E: Exception do
      begin
        ShowMessage(E.Message);
        Exit;
      end;
    end;
  finally
    Parser.Free;
  end;

  StartPos := Provider.StartPosition;
  Text := Provider.FullText;

  Index := Length(Text);

  if Text[Index] = ')' then
    Dec(Index);
  while CharInSet(Text[Index], [#0..#31]) do
    Dec(Index);

  WriteSimpleText(ParserOutput, Index, StartPos);
end;

{ TCharProvider }

function TCharProvider.NextBuffer: Boolean;
var
  Buffer: array[0..1024] of AnsiChar;
  BufferLength: Integer;
  Editor: IOTASourceEditor;
  Reader: IOTAEditReader;
begin
  Editor := EditorTools.GetEditor;
  Reader := Editor.CreateReader;
  BufferLength := Reader.GetText(FReadPos, Buffer, Length(Buffer) - 1);
  Buffer[1024] := #0;
  Result := BufferLength > 0;
  if Result then
  begin
    Inc(FReadPos, BufferLength);
    FFullText := FFullText + string(Buffer);
  end;
end;

procedure TCharProvider.Back;
begin
  if FStringPos > 1 then
    Dec(FStringPos);
end;

constructor TCharProvider.Create(AtCurrentPos: Boolean = False);
var
  Index: Integer;
  Start: Integer;
  Line: string;
begin
  Line := GetCurrentLine(EditorTools.GetEditor, Start, Index);
  if AtCurrentPos then
  begin
    if Line <> '' then
    begin
      while (Index > 1) and
            CharInSet(Line[Index - 1], ['A'..'Z', 'a'..'z', '0'..'9', '_']) do
        Dec(Index);
      while (Index < System.Length(Line)) and
            not CharInSet(Line[Index], ['A'..'Z', 'a'..'z', '0'..'9', '_']) do
        Inc(Index);
    end;
    FStartPos := Start + Index - 1;
  end
  else
    FStartPos := Start;
  FReadPos := FStartPos;
  FStringPos := 1;
  NextBuffer;
end;

function TCharProvider.CurrentPosition: Integer;
begin
  Result := FStartPos + FStringPos;
end;

function TCharProvider.FullText: string;
begin
  Result := Copy(FFullText, 1, FStringPos - 1);
end;

function TCharProvider.LookAhead: Char;
begin
  if FStringPos > Length(FFullText) then
    if NextBuffer then
      Result := FFullText[FStringPos]
    else
      Result := CLexers.LexerEndMarker
  else
    Result := FFullText[FStringPos];
end;

function TCharProvider.NextChar: Char;
begin
  Result := LookAhead;
  Inc(FStringPos);
end;

function TCharProvider.StartPosition: Integer;
begin
  Result := FStartPos;
end;

end.
