unit EditorStrings;

interface

uses Classes, Contnrs, ToolsAPI;

type
  TEditorStrings = class(TStrings)
  private
    FEditor: IOTASourceEditor;
    FEditView: IOTAEditView;
    FObjects: TObjectList;
    function LineLength(Index: Integer): Longint;
    function GetLinePos(Index: Integer): Longint;
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
  public
    constructor Create(Editor: IOTASourceEditor);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
  end;

implementation

uses RTLConsts;

{ TEditorStrings }

// Index is 0-based, result value is offset into editor buffer
function TEditorStrings.GetLinePos(Index: Integer): Longint;
var
  CharPos: TOTACharPos;
begin
  CharPos.CharIndex := 0;
  CharPos.Line := Index + 1;
  Result := FEditView.CharPosToPos(CharPos);
end;

// Includes a possible LineBreak
function TEditorStrings.LineLength(Index: Integer): Longint;
begin
  Result := GetLinePos(Index + 1) - GetLinePos(Index);
end;

procedure TEditorStrings.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TEditorStrings then
  begin
    FObjects.Assign(TEditorStrings(Source).FObjects);
    FEditor := TEditorStrings(Source).FEditor;
    FEditView := TEditorStrings(Source).FEditView;
  end;
end;

procedure TEditorStrings.Clear;
var
  Writer: IOTAEditWriter;
begin
  inherited;
  Writer := FEditor.CreateUndoableWriter;
  try
    Writer.DeleteTo(GetLinePos(FEditor.GetLinesInBuffer));
  finally
    Writer := nil;
  end;
end;

constructor TEditorStrings.Create(Editor: IOTASourceEditor);
begin
  inherited Create;
  FEditor := Editor;
  FEditView := Editor.EditViews[0];
  FObjects := TObjectList.Create;
  while FObjects.Count < GetCount do
    FObjects.Add(nil);
end;

procedure TEditorStrings.Delete(Index: Integer);
var
  Writer: IOTAEditWriter;
begin
  Writer := FEditor.CreateUndoableWriter;
  try
    Writer.CopyTo(GetLinePos(Index));
    Writer.DeleteTo(GetLinePos(Index + 1));
  finally
    Writer := nil;
  end;
  FObjects.Delete(Index);
end;

destructor TEditorStrings.Destroy;
begin
  FObjects.Free;
  inherited;
end;

function TEditorStrings.Get(Index: Integer): string;
var
  Reader: IOTAEditReader;
begin
  if (Index < 0) or (Index >= Count) then
    Error(@SListIndexError, Index);
  SetLength(Result, LineLength(Index));
  Reader.GetText(GetLinePos(Index), PChar(Result), LineLength(Index));
  if Pos(LineBreak, Result) = Length(Result) - Length(LineBreak) + 1 then
    SetLength(Result, Length(Result) - Length(LineBreak));
end;

function TEditorStrings.GetCount: Integer;
begin
  Result := FEditor.GetLinesInBuffer;
end;

function TEditorStrings.GetObject(Index: Integer): TObject;
begin
  while FObjects.Count < GetCount do
    FObjects.Add(nil);
  Result := FObjects[Index];
end;

procedure TEditorStrings.Insert(Index: Integer; const S: string);
var
  Writer: IOTAEditWriter;
begin
  if (Index < 0) or (Index > GetCount) then
    Error(@SListIndexError, Index);
  Writer := FEditor.CreateUndoableWriter;
  try
    Writer.CopyTo(GetLinePos(Index));
    Writer.Insert(PChar(S + LineBreak));
  finally
    Writer := nil;
  end;
  FObjects.Insert(Index, nil);
end;

procedure TEditorStrings.PutObject(Index: Integer; AObject: TObject);
begin
  while FObjects.Count < GetCount do
    FObjects.Add(nil);
  FObjects[Index] := AObject;
end;

end.
