{                                                                           }
{ File:       TextMan.pas                                                   }
{ Function:   Parser helper code                                            }
{ Language:   Delphi for Win32                                              }
{ Author:     Rudy Velthuis                                                 }
{ Copyright:  (c) 2005 Rudy Velthuis                                        }
{ Disclaimer: This code is freeware. All rights are reserved.               }
{             This code is provided as is, expressly without a warranty     }
{             of any kind. You use it at your own risk.                     }
{                                                                           }
{             If you use this code, please credit me.                       }
{                                                                           }

unit TextMan;

// This unit is a very simple parser that considers anything
// delimited by white space characters a symbol.

interface

uses
  SysUtils;

const
  // Default character sets.
  CWhiteSpaceDef  = [#9, #10, #13, ' '];
  CIdentifierDef  = ['A'..'Z', 'a'..'z', '0'..'9', '_', ':']; // Pascal/C ID
  CParDef         = ['{', '$', '}'];                     // .PAR file specifics
  CExternalSymDef = CIdentifierDef + CParDef;

  CRLF            = #13#10;

var
  // Allow user to change character sets.
  WhiteSpaceChars: TSysCharSet = CWhiteSpaceDef;
  IdentifierChars: TSysCharSet = CIdentifierDef;
  ParChars: TSysCharSet = CParDef;
  ExternalSymChars: TSysCharSet = CExternalSymDef;

function SymbolStart(const Line: string; Index: Integer): Integer;
function SymbolLength(const Line: string; Index: Integer): Integer;
function IdentifierStart(const Line: string; Index: Integer): Integer;
function IdentifierLength(const Line: string; Index: integer): Integer;
function NextSymbolStart(const Line: string; Index: Integer): Integer;
function NextSymbolStartP(const Line: string; Index: Integer; var NumPtrs: Integer): Integer;
function SkipIdentifierP(const Line: string; Index: Integer; var NumPtrs: Integer): Integer;
function SymbolAtIndex(const Line: string; Index: Integer): string;
function IdentifierAtIndex(const Line: string; Index: Integer): string;
function LeadingWhiteSpace(const Line: string): string;
function SkipWhiteSpace(Line: string; Index: Integer): Integer;
function SkipComment(Line: string; var Index: Integer): Boolean;
function DecodeOctal(Value: string): Integer;

implementation

{$IFNDEF UNICODE}
uses
  UnicodeUtils;
{$ENDIF}

function SymbolAtIndex(const Line: string; Index: Integer): string;
var
  SymLen: Integer;
begin
  Index := SymbolStart(Line, Index);
  SymLen := SymbolLength(Line, Index);
  if SymLen > 0 then
    Result := Copy(Line, Index, SymLen)
  else
    Result := '';
end;

function IdentifierAtIndex(const Line: string; Index: Integer): string;
var
  IdLen: Integer;
begin
  Index := IdentifierStart(Line, Index);
  IdLen := IdentifierLength(Line, Index);
  if IdLen > 0 then
    Result := Copy(Line, Index, IdLen)
  else
    Result := '';
end;

function LeadingWhiteSpace(const Line: string): string;
var
  Len, IndentLen: Integer;
begin
  Len := Length(Line);
  IndentLen := 0;
  while (IndentLen < Len) and CharInSet(Line[1 + IndentLen], WhiteSpaceChars) do
    Inc(IndentLen);
  if IndentLen > 0 then
    Result := Copy(Line, 1, IndentLen)
  else
    Result := '';
end;

function SymbolStart(const Line: string; Index: Integer): Integer;
begin

  // If on white space, go to first non-white-space.
  SkipWhiteSpace(Line, Index);

  // While on non-white-space go back to beginning of symbol.
  while (Index > 1) and not CharInSet(Line[Index - 1], WhiteSpaceChars) do
    Dec(Index);

  Result := Index;
end;

function SymbolLength(const Line: string; Index: Integer): Integer;
var
  Len: Integer;
begin
  Len := Length(Line);
  Index := SymbolStart(Line, Index);

  // Find first white space after symbol start or end of line
  // (which comes first).
  Result := Index;
  while (Result < Len)  and
        not CharInSet(Line[Result], WhiteSpaceChars) do
    Inc(Result);
  Dec(Result, Index);
end;

function NextSymbolStart(const Line: string; Index: Integer): Integer;
begin
  Index := SymbolStart(Line, Index) + SymbolLength(Line, Index);
  if Index >= Length(Line) then
    Result := Index
  else
    Result := SkipWhiteSpace(Line, Index);
end;

function NextSymbolStartP(const Line: string; Index: Integer; var NumPtrs: Integer): Integer;
begin
  Index := SymbolStart(Line, Index) + SymbolLength(Line, Index);
  if Index >= Length(Line) then
    Result := Index
  else
  begin
    while CharInSet(Line[Index], WhiteSpaceChars + ['*']) do
    begin
      if Line[Index] = '*' then
        Inc(NumPtrs);
      Inc(Index);
    end;
    Result := Index;
  end;
end;

function SkipIdentifierP(const Line: string; Index: Integer; var NumPtrs: Integer): Integer;
begin
  Index := IdentifierStart(Line, Index) + IdentifierLength(Line, Index);
  if Index >= Length(Line) then
    Result := Index
  else
  begin
    while CharInSet(Line[Index], WhiteSpaceChars + ['*']) do
    begin
      if Line[Index] = '*' then
        Inc(NumPtrs);
      Inc(Index);
    end;
    Result := Index;
  end;
end;


function IdentifierStart(const Line: string; Index: Integer): Integer;
begin

  // If on white space, go to first non-white-space.
  Index := SkipWhiteSpace(Line, Index);

  // While on non-white-space go back to beginning of symbol.
  while (Index > 1) and CharInSet(Line[Index - 1], IdentifierChars) do
    Dec(Index);

  Result := Index;
end;

function IdentifierLength(const Line: string; Index: Integer): Integer;
var
  I: Integer;
begin
  I := IdentifierStart(Line, Index);
  while CharInSet(Line[I], IdentifierChars) do
    Inc(I);
  Result := I - Index;
end;

// TODO : Find a way to keep the comment(s).
//        Perhaps each comment character can be stored in a string. 
function SkipComment(Line: string; var Index: Integer): Boolean;
var
  P: PAnsiChar;
begin
  Result := True;
  P := @Line[Index];
  if P^ = '/' then
  begin
    Inc(P);
    case P^ of
      '/':
        begin
          while not (P^ in [#0, #13]) do
            Inc(P);
          if (P^ = #13) and ((P + 1)^ = #10) then
            Inc(P, 2);
        end;
      '*':
        begin
          Inc(P);
          while (P <> PAnsiChar(#0)) and not ((P^ = '*') and (P[1] = '/')) do
            Inc(P);
          if P^ <> #0 then
            Inc(P, 2);
        end;
    end;
    Result := P^ <> #0;
    Index := P - @Line[1] + 1;
  end;
end;

function SkipWhiteSpace(Line: string; Index: Integer): Integer;
begin
  while (Index <= Length(Line)) do
    if Line[Index] = '/' then
      SkipComment(Line, Index)
    else if CharInSet(Line[Index], WhiteSpaceChars) then
      Inc(Index)
    else
      Break;
  Result := Index;
end;

function DecodeOctal(Value: string): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 1 to Length(Value) do
    Result := 8 * Result + (Ord(Value[I]) - Ord('0'));
end;

end.
